from margument.log import throw


class SettingsProcessor:
    def __init__(self, settings, args_parser):
        """
        :param settings: A list of settings to be processed.
        :param args_parser: the instance from argsparse.ArgumentParser() previously created
        """
        self.settings = settings
        self.args_parser = args_parser

    def run(self):
        """
        This method will iterate through all the given settings and will process each one according to their specifications.
        :return: A dictionary of all settings given each containing a list of all parsed configurations from the user and/or the external file .
        """
        for setting in self.settings:
            setting.program_arguments.add_arguments(self.args_parser)

        user_arguments = None
        try:
            user_arguments = self.args_parser.parse_args()
        except AssertionError as ex:
            throw(ex)

        parsed_settings = {}
        for setting in self.settings:
            setting.user_arguments = user_arguments

            setting.program_arguments.process_arguments(self.settings)

            setting.do_commands()

            setting.set()

            parsed_settings[setting.program_arguments.__class__.__name__] = setting.save()

        return parsed_settings



class Argument:
    def __init__(self,
                 name='',
                 abbreviation_name='',
                 full_name='',
                 value=None,
                 help_message='',
                 metavar='',
                 default='None',
                 to_save=False,
                 is_main=False,
                 command=None,
                 command_args=None,
                 choices=None):
        """
        :param name: the name of the argument, ex: alias (must be the same as full_name)
        :param abbreviation_name: the abbreviation name for the argument, ex: -a (see argsparse documentation https://docs.python.org/3/library/argparse.html#name-or-flags)
        :param full_name: the full name of the argument, ex: --alias (must be the same as name) (see argsparse documentation https://docs.python.org/3/library/argparse.html#name-or-flags)
        :param value: the value of the argument provided posteriorly by the user.
        :param help_message: the help message for the argument (see argsparse documentation https://docs.python.org/3/library/argparse.html#help)
        :param metavar: the metavar for the argument (see argsparse documentation https://docs.python.org/3/library/argparse.html#metavar)
        :param default: the default value for the argument
        :param to_save: a boolean defining if this argument should be saved in the yaml file.
        :param is_main: a boolean defining if this argument is a main argument. There must be 1 if to use on RepeatableSettings class
        :param command: the method that this argument should call when it's present on the arguments parsed and provided by the user.
        :param command_args: all the arguments the previous method needs to have. If more than one argument, please wrap them in (), ex: (arg1, arg2)
        :param choices: a list of choices required by the argument (see argsparse documentation https://docs.python.org/3/library/argparse.html#choices)
        """
        self.name = name
        self.abbreviation_name = abbreviation_name
        self.full_name = full_name
        self.value = value
        self.help_message = help_message
        self.metavar = metavar
        self.default = default
        self.to_save = to_save
        self.is_main = is_main
        self.command = command
        self.command_args = command_args
        self.choices = choices

    @staticmethod
    def set_argument(argument):
        return Argument(
            name=argument.name,
            abbreviation_name=argument.abbreviation_name,
            full_name=argument.full_name,
            value=argument.value,
            help_message=argument.help_message,
            metavar=argument.metavar,
            default=argument.default,
            to_save=argument.to_save,
            is_main=argument.is_main,
            command=argument.command,
            command_args=argument.command_args,
            choices=argument.choices)

    def set_value(self, value):
        self.value = value

    def set_command(self, command):
        self.command = command

    def set_command_args(self, command_args):
        """
        If more than one argument, please wrap them in (), ex: (arg1, arg2)
        :param command_args: all the arguments required by the method/command
        :return: command the arguments defined
        """
        self.command_args = command_args



class Options:

    def __init__(self, show_saved=False, save_different=False, save_main_arg_exists=False, custom_save=False):
        """
        :param show_saved: If True, it will show all saved configuration in the terminal, otherwise False. Default: False.
        :param save_different: If True, it will only save the configurations if they are different from the configurations already in the file. If
        False, it will not save. Default: False.
        :param save_main_arg_exists: If True, it will only save the configurations if the main argument is present in the arguments provided by the user. If
        False, it will not save. Default: False.
        :param custom_save: True or False depending on any custom condition. If True it saves, if False it will not save. Default: False.
        """
        self.show_saved = show_saved
        self.save_different = save_different
        self.save_main_arg_exists = save_main_arg_exists
        self.custom_save = custom_save




import yaml


def write_yaml(path, configs):
    config_file = open(path, 'w')
    yaml.safe_dump(configs, config_file)
    config_file.close()


def read_yaml(path):
    config_file = open(path, 'r')
    configs = yaml.safe_load(config_file)
    config_file.close()
    return configs



import os.path

from margument.options import Options
from margument.file import File
from margument.yaml import read_yaml
from margument.log import show


class Settings:

    def __init__(self, path, program_arguments=None, options=Options()):
        self.file = File(path)
        self.program_arguments = program_arguments
        self.settings_from_file = {}
        self.values_to_save = {}
        self.user_arguments = None
        self.configs = {}
        self.options = options

        if self.exists():
            self.load()

    def set(self):
        raise NotImplementedError("This method needs to be implemented.")

    def save(self):
        raise NotImplementedError("This method needs to be implemented.")

    def do_commands(self):
        non_saved_arguments = self.program_arguments.to_list(is_saved=[False])
        for arg in non_saved_arguments:
            if arg.command is None:
                continue

            if arg.name in self.user_arguments:
                if arg.command_args is not None:
                    if isinstance(arg.command_args, tuple):
                        arg.command(*arg.command_args)
                    else:
                        arg.command(arg.command_args)
                else:
                    arg.command()

    def load(self):
        self.settings_from_file = read_yaml(self.file.path)

    def exists(self):
        if not os.path.exists(self.file.path):
            return False

        if self.file.is_empty():
            return False
        return True

    def set_arguments_values(self, values_from_file):
        program_arguments = self.program_arguments.__class__()
        arguments = program_arguments.to_list()

        for configuration in arguments:
            self.set_settings_value(configuration, values_from_file)

        program_arguments.from_list(arguments)

        return program_arguments

    def set_settings_value(self, config, values_from_file):
        raise NotImplementedError("This method needs to be implemented.")

    def save_when_different(self):
        if not self.options.save_different:
            return self.options.save_different

        if self.settings_from_file == self.values_to_save:
            return False
        return True

    def show(self, arguments):
        message = 'Saved configurations:'
        for name, value in arguments.items():
            message += '\n\t\t\t' + name + ': ' + str(value)

        show(message + '\n\n\t\tConfiguration file path: ' + self.file.path)



import sys


def throw(message):
    print('\t>>!ERROR:\n\t\t' + message)
    sys.exit()


def show(message, to_exit=False):
    print('    >>!:\n\t\t' + message + '\n')

    if to_exit:
        sys.exit()



from margument.options import Options
from margument.settings import Settings
from margument.yaml import write_yaml


class NonRepeatableSettings(Settings):

    def __init__(self, path, program_arguments=None, options=Options()):
        super().__init__(path, program_arguments, options)

    def set(self):
        self.configs = self.set_arguments_values(self.settings_from_file)

    def set_settings_value(self, config, values_from_file):
        if config.name in self.user_arguments:
            value = getattr(self.user_arguments, config.name)
        elif config.name in self.settings_from_file:
            value = self.settings_from_file[config.name]
        else:
            value = config.default

        config.set_value(value)

        if config.to_save:
            self.values_to_save[config.name] = config.value

    def save(self):
        if self.save_when_different() or self.options.custom_save:
            write_yaml(self.file.path, self.values_to_save)

            if self.options.show_saved:
                self.show(self.values_to_save)

        return self.configs



class File:
    file_in_use = None
    __lines = []

    def __init__(self, path):
        self.path = path

    def open(self, mode):
        self.file_in_use = open(self.path, mode)
        return self.file_in_use

    def close(self):
        self.file_in_use.close()

    def write(self, line):
        self.file_in_use.write(line)

    def write_lines(self, lines):
        self.file_in_use.writelines(lines)
        self.close()

    def set_lines(self):
        self.__lines = open(self.path).readlines()

    def get_lines(self):
        return self.__lines

    def is_empty(self):
        self.set_lines()
        return len(self.get_lines()) == 0



from setuptools import setup
import pathlib
from margument.version.progsettings import get_version

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

version = get_version()

setup(
    name="margument",
    version=version,
    description="python library to manage configurations from program arguments including doing commands and saving configurations in a yaml file.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zaytiri/settings-manager",
    project_urls={
        'GitHub': 'https://github.com/zaytiri/settings-manager',
        'Changelog': 'https://github.com/zaytiri/settings-manager/blob/main/CHANGELOG.md',
    },
    author="zaytiri",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    keywords="manager, configurations, settings, arguments, argparse, yaml, save",
    package_data={'margument': ['progsettings.yaml']},
    packages=["margument", "margument.version"],
    python_requires=">=3.10.6",
    install_requires=[
      "PyYAML~=6.0"
    ],
    # entry_points={
    #     "console_scripts": [
    #         "margument=margument:app.main",
    #     ],
    # }
)


from margument.log import throw
from margument.options import Options
from margument.reflection import convert_to_dict
from margument.settings import Settings
from margument.yaml import write_yaml


class RepeatableSettings(Settings):
    __main_arg = {'name': '', 'value': ''}

    def __init__(self, path, program_arguments=None, options=Options()):
        super().__init__(path, program_arguments, options)

    def __process_main_arg(self):
        main_arg_exists = False
        for arg in self.program_arguments.to_list():
            if arg.is_main:
                main_arg_exists = True
                self.__main_arg['name'] = arg.name
                break

        if not main_arg_exists:
            throw("No main argument defined.")

        if self.__main_arg['name'] in self.user_arguments:
            self.__main_arg['value'] = getattr(self.user_arguments, self.__main_arg['name'])

    def set(self):
        self.__process_main_arg()

        user_arguments_dict = convert_to_dict(self.user_arguments)

        if not self.exists():
            try:
                self.configs[self.__main_arg['value']] = self.set_arguments_values(user_arguments_dict)
            except KeyError:
                pass
            return

        for saved_setting in self.settings_from_file:
            self.configs[saved_setting] = self.set_arguments_values(self.settings_from_file[saved_setting])

        if self.__main_arg['name'] not in self.user_arguments:
            return

        if self.__main_arg['value'] not in self.configs:
            self.configs[self.__main_arg['value']] = self.set_arguments_values(user_arguments_dict)

    def set_settings_value(self, config, values_from_file):
        is_main = False
        if self.__main_arg['value'] != '':
            is_main = values_from_file[self.__main_arg['name']] == self.__main_arg['value']

        if config.name in self.user_arguments and is_main:
            value = getattr(self.user_arguments, config.name)
        else:
            if config.name in values_from_file:
                value = values_from_file[config.name]
            else:
                value = config.default

        config.set_value(value)
        if config.to_save and is_main:
            self.values_to_save[config.name] = config.value

    def save_when_main_arg_exist(self):
        if not self.options.save_main_arg_exists:
            return self.options.save_main_arg_exists

        if self.__main_arg['name'] in self.user_arguments:
            return True
        return False

    def save(self):
        if self.save_when_different() or self.save_when_main_arg_exist() or self.options.custom_save:
            self.settings_from_file[self.__main_arg['value']] = self.values_to_save

            write_yaml(self.file.path, self.settings_from_file)

            if self.options.show_saved:
                self.show(self.values_to_save)

        return self.configs



from margument.argument import Argument
from margument.reflection import get_class_variables


class Arguments:
    def to_list(self, is_saved=None):
        """
        It will convert all properties from this class instance into a list.
        :param is_saved: it will convert to list all properties that are set to be saved if True, otherwise if False. If None, it will convert all
        properties regardless.
        :return: a list containing all properties of this class instance.
        """
        if is_saved is None:
            is_saved = [True, False]

        class_variables = get_class_variables(self)

        arguments = []
        for var in class_variables:
            if not isinstance(var[1], Argument):
                continue

            arg = Argument.set_argument(var[1])

            if arg.to_save not in is_saved:
                continue

            arguments.append(arg)

        return arguments

    def from_list(self, arguments):
        """
        It will convert all list elements into their respective property in this class instance.
        :param arguments: a list from to_list() method
        """
        class_variables = get_class_variables(self)

        for var in class_variables:
            if not isinstance(var[1], Argument):
                continue

            arg = Argument.set_argument(var[1])

            arg_from_list = None
            for argument in arguments:
                if argument.name != arg.name:
                    continue
                arg_from_list = argument

            if arg_from_list is None:
                continue

            setattr(self, arg.name, arg_from_list)

    def add_arguments(self, args_parser):
        """
        All argument configuration from argsparse must be defined in this method. (see argsparse documentation https://docs.python.org/3/library/argparse.html#the-add-argument-method)
        :param args_parser: the instance from argsparse.ArgumentParser()
        """
        raise NotImplementedError("This method needs to be implemented.")

    def process_arguments(self, settings):
        """
        All processing of parsed user defined arguments from the argsparse must be done in this method. Processing examples: validation of fields
        from the parsed arguments from argsparse, validation of the external yaml file or definition of necessary command arguments.
        :param settings:
        """
        raise NotImplementedError("This method needs to be implemented.")